// At 3.0V, ~830nA average. At 1.8V, ~495nA, trickle charge off

// PIC16LF1947 Configuration Bit Settings

// 'C' source line config statements

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#ifdef __DEBUG
    #pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#else
    #pragma config PWRTE = ON       // Power-up Timer Enable (PWRT enabled)
#endif
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Memory Code Protection (Data memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover (Internal/External Switchover mode is disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is disabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config PLLEN = OFF      // PLL Enable (4x PLL disabled)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

#include <xc.h>
#include <stdint.h>
#include <stdbool.h>
#include "lcd.h"

// #define ENABLE_VBAK_TRICKLE_CHARGE // Adds ~45nA @ 1.8V. If disabled, tie Vbsk to Vss with a 10k resistor.

#define _XTAL_FREQ (16000000UL)

#define SDA_TRIS TRISEbits.TRISE0
#define SCL_TRIS TRISEbits.TRISE1
#define SDA_ANSEL ANSELEbits.ANSE0
#define SCL_ANSEL ANSELEbits.ANSE1
#define SDA_LAT LATEbits.LATE0
#define SCL_LAT LATEbits.LATE1
#define SDA_PORT PORTEbits.RE0

#define SDA_DIGITAL()
#define SCL_DIGITAL()
#define SDA_HIGH() SDA_TRIS=1
#define SDA_LOW() SDA_TRIS=0
#define SDA_INPUT() SDA_TRIS=1
#define SDA_OUTPUT() SDA_TRIS=0
#define SCL_HIGH() SCL_TRIS=1
#define SCL_LOW() SCL_TRIS=0
#define SCL_INPUT() SCL_TRIS=1
#define SCL_OUTPUT() SCL_TRIS=0

#define I2C_RV3028C7_ADDR (0x52)

// The TIMER1 delay is measured for 1024Hz clock (2048Hz from T1CLKI/RTC CLKOUT plus 1:2 prescaler).
//#define TIMER1_DELAY (1024U*60U) // 1 minute
//#define TIMER1_DELAY (1024U) // 1 second, for testing
#define TIMER1_DELAY (1024U/2U) // 0.5 second

static const char _szBuildTime[]=__TIME__; // Used to reset the RTC at first time boot, format "HH:MM:SS"
//static const char _szBuildTime[]="11:11:11"; // Used to reset the RTC at first time boot, format "HH:MM:SS"
//static const char _szBuildTime[]="01:22:22"; // Used to reset the RTC at first time boot, format "HH:MM:SS": 01:22 is 18 segs, average for 4 digit display; 01:22:22 is 28 segs, average for 6 digit display

typedef struct
{
    int8_t n8Hours10;
    int8_t n8Hours1;
    int8_t n8Mins10;
    int8_t n8Mins1;
    int8_t n8Secs10;
    int8_t n8Secs1;
    int8_t n8HalfSecond;
} CLKSTRUCT;

static CLKSTRUCT _cs;

typedef union
{
    uint8_t u8;
    struct
    {
        unsigned b0                    :1;
        unsigned b1                    :1;
        unsigned b2                    :1;
        unsigned b3                    :1;
        unsigned b4                    :1;
        unsigned b5                    :1;
        unsigned b6                    :1;
        unsigned b7                    :1;
    } s;
} U8UNION;

static void I2CStart(void)
{
    SDA_HIGH();
    SCL_HIGH();
    __delay_us(3);
    SDA_LOW();
    __delay_us(3);
    SCL_LOW();                        // pull SCL low
}

static void I2CStop(void)
{
    SCL_LOW();                        // ensure SCL is low
    __delay_us(3);
    SDA_LOW();                        // SDA low
    __delay_us(3);
    SCL_HIGH();                        // pull SCL high
    __delay_us(3);
    SDA_HIGH();                   // allow SDA to be pulled high
    __delay_us(3);
}

static void I2CWrite(uint8_t u8Data)
{
    int8_t i=8;                // loop counter

    do         // loop through each bit
    {
        if (((U8UNION)u8Data).s.b7)
        {
            SDA_HIGH();
        }
        else
        {
            SDA_LOW();
        }
        // Data setup time is 250ns min
        SCL_HIGH();             // pull SCL high to clock bit
        u8Data <<= 1;           // shift left for next bit - Do it here to avoid additional delays when SCL=0
        NOP(); // Pad out SCL high to min 4us
        NOP(); // Pad out SCL high to min 4us
        SCL_LOW();                        // pull SCL low for next bit            
    }
    while (--i!=0); // Hopefully doing it this way XC8 will use DECFSZ instruction
    
    // Check for ACK
    SDA_HIGH();                    // configure SDA as an input
    // No need for delay @ Fosc=4MHz
    SCL_HIGH();                        // bring SCL high to begin transfer
    while (SDA_PORT) // Wait for ACK
    {
    }
    SCL_LOW();                        // bring SCL low again.    
}

static uint8_t I2CRead(bool bNack)
{
    int8_t i=8;
    uint8_t u8Data=0;

    SDA_HIGH();                   // configure SDA as an input

    do 
    {
        u8Data <<= 1;
        SCL_HIGH();                        // bring SCL high to begin transfer
        if (SDA_PORT)
        {
            u8Data|=1;
        }
        SCL_LOW();                        // bring SCL low again.     
    }
    while (--i!=0);

    if (bNack)
    {
        SDA_HIGH();
    }
    else
    {
        SDA_LOW();
    }
    // Data setup time is 250ns min
    SCL_HIGH();                       // pull SCL high to clock bit
    __delay_us(3); 
    SCL_LOW();                        // pull SCL low for next bit
    return u8Data;
}

static void I2CWriteRegister(uint8_t u8I2CAddr,uint8_t u8RegAddr,uint8_t u8Data)
{
	I2CStart();
	
	I2CWrite((uint8_t)(u8I2CAddr << 1) | 0U);
	I2CWrite(u8RegAddr);
	I2CWrite(u8Data);
	I2CStop();
}

uint8_t I2CReadRegister(uint8_t u8I2CAddr,uint8_t u8RegAddr)
{
  	uint8_t u8Data;
  
	I2CStart();
	
	I2CWrite((uint8_t)(u8I2CAddr << 1) | 0U);
	I2CWrite(u8RegAddr);
	I2CStart();
	I2CWrite((uint8_t)(u8I2CAddr << 1) | 1U);
	u8Data=I2CRead(true);
	I2CStop();
            
   	return u8Data;
} 

static void I2CInit(void)
{
    // I2C bit bang uses GPIOs in simulated open drain mode
    SDA_DIGITAL();
    SCL_DIGITAL();
    SDA_INPUT();
    SCL_INPUT();
    SDA_LAT=0;
    SCL_LAT=0;
#ifdef SDA_ANSEL
    SDA_ANSEL=0;
#endif    
#ifdef SCL_ANSEL
    SCL_ANSEL=0;
#endif    
}

static uint8_t AsciiBCDToPackedBCD(const char *psz)
{
    uint8_t u8=((uint8_t)psz[1]) & 0x0F;
    
    u8 |= (((uint8_t)psz[0])<<4);
    
    return u8;
}

static void LPDelayMS(uint8_t u8Delay)
{
    // Low power delay: switch to LFINTOSC 31kHz
    // Note that Timer 0 will not operate in sleep so we have to poll
    // 255 >= u8Delay >= 1
    // delay[ms] = 2.55 + (0.956 * u8MX) empirically approx, or:
    // u8Delay = (delay[ms] - 2.55) / 0.956 approx
    
    uint8_t u8IRCFSave=OSCCONbits.IRCF;
    
    OPTION_REGbits.TMR0CS=0; // 0 => Fosc/4
    OPTION_REGbits.PS=0b010; // 0b010 => 1:8 timer 0 prescaler ( @ 31kHz Fosc, ~= 1ms [ Fosc/4/8] )
    OPTION_REGbits.PSA=0; // 0 => Prescaler is assigned to the Timer0 module
    INTCONbits.TMR0IE=0;
    OSCCONbits.IRCF=0b0000; // 0b0000 => LFINTOSC
    while (!OSCSTATbits.LFIOFR) // Wait for LFINTOSC to be ready
    {
    }
    TMR0=(uint8_t)(256-u8Delay);
    INTCONbits.TMR0IF=0;
    while (!TMR0IF)
    {
    }
    
    OSCCONbits.IRCF=u8IRCFSave;
    OPTION_REGbits.TMR0CS=1; // 1 => T0CKI (POR default))
    OPTION_REGbits.PS=0b11; // 0b111 => 1:256 timer 0 prescaler (POR default)
    OPTION_REGbits.PSA=1; // 0 => Prescaler is not assigned to the Timer0 module (default)
    TMR0=0;
}

static void RTCInit(void)
{
    // Wait for RTC's EEbusy to go inactive
    {
        bool bWaitingForRTC=true;

        while (bWaitingForRTC)
        {
            uint8_t u8RTCStatus=I2CReadRegister(I2C_RV3028C7_ADDR,0x0E); // Read RTC status register
            
            bWaitingForRTC=(u8RTCStatus & 0x80)!=0; // Check EEbusy bit
            if (bWaitingForRTC)
            {
                LPDelayMS(50); // Low power delay
            }
        }
    }
    
    // Use the build date & time for our initial time
    {
        uint8_t u8;

        // Convert __TIME__ format HH:MM:SS to BCD digits, one nibble per byte, into RAM clock
        u8=AsciiBCDToPackedBCD(_szBuildTime+6);
        _cs.n8Secs10 = u8 >> 4;
        _cs.n8Secs1 = u8 & 0x0F;
        
        u8=AsciiBCDToPackedBCD(_szBuildTime+3);
        _cs.n8Mins10 = u8 >> 4;
        _cs.n8Mins1 = u8 & 0x0F;

        u8=AsciiBCDToPackedBCD(_szBuildTime+0);
        _cs.n8Hours10 = u8 >> 4;
        _cs.n8Hours1 = u8 & 0x0F;
        
        _cs.n8HalfSecond=0;
    }
    
    // Put 2048Hz on CLKOUT pin (wired to T1CKI))
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x0A,0x02); // Timer value 0
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x0B,0x00); // Timer value 1
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x0F,0b10011100); // Control 1 register; Periodic countdown enabled[TE-2], [TRPT-7]repeated, [TD-0:1] 4096Hz clock source, [USEL-4] set, [EERD-3] EEPROM daily refresh disabled
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x10,0b00010000); // Control 2 register; [UIE-5] disabled, [TIE-4] enabled

    I2CWriteRegister(I2C_RV3028C7_ADDR,0x35,0xC6); // Periodic countdown on CLKOUT, wired to T1CKI

#ifdef ENABLE_VBAK_TRICKLE_CHARGE // Adds ~45nA @ 1.8V
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x37,0x37); // TCE[5] on; FEDE[4] on (default); BSM[3:2]=0b01, DSM; TCR[0:1]=0b11 => 15k trickle charge; DSM adds 45nA @1.8V
#else        
    I2CWriteRegister(I2C_RV3028C7_ADDR,0x37,0x10); // TCE[5] off; FEDE[4] on (default); TCR[0:1]=0b00 => 3k trickle charge ~(default)
#endif

    I2CWriteRegister(I2C_RV3028C7_ADDR,0x0E,0x00); // Clear any outstanding POR in RTC status register
}

static void CPUInit(void)
{
    OSCCONbits.IRCF=0b1111; // 0b1111 => 16MHz
    
    // GPIO direction
    TRISA=0b00000000; 
    TRISB=0b00000000;
    TRISC=0b00000001; // RC0/T1CKI
    TRISD=0b00000000;
    TRISE=0b00000111; // RE2 is VLCD3, RE1 is SCL, RE0 is SDA
    TRISF=0b00000000;
    TRISG=0b00100000; // RG5 is MCLR
    ANSELA=0;
    ANSELE=0;
    ANSELF=0;
    ANSELG=0;
}

static void T1OscInit(void) // Enable T1CKI, it's required for LCD, wired the RTC's CLKOUT.
{
    TRISCbits.TRISC0=1; // RC0/T1CKI. No analog function on this pin, so no ANSEL bit config to perform
    T1CONbits.T1OSCEN=0;   
}

static void T1Init(void) // Setup timer 1 for 1 second interrupt: T1CKI should already have 2048Hz on it. T1OscInit() should have been run before this.
{
    T1CONbits.TMR1CS=0b10; // 0b10 => T1CKI
//    T1CONbits.T1CKPS=0b00; // 0b00 => 1:1 prescale
    T1CONbits.T1CKPS=0b01; // 0b01 => 1:2 prescale => 1024Hz TMR1 increment from 2048Hz T1CLKI
    T1CONbits.nT1SYNC=1; // 1 => Not sync'd with Fcy 
    TMR1=(uint16_t)(65536UL-TIMER1_DELAY-1); // 1min timeout on timer 1
    PIR1bits.TMR1IF=0;
    PIE1bits.TMR1IE=1;
    T1CONbits.TMR1ON=1;
}

static void LCDUpdate(void)
{
    // Setup the display digits with the current RAM time
    LCDDigit1Set((uint8_t)_cs.n8Hours10);
    LCDDigit2Set((uint8_t)_cs.n8Hours1);
    LCDDigit3Set((uint8_t)_cs.n8Mins10);
    LCDDigit4Set((uint8_t)_cs.n8Mins1);    
    LCDDigit5Set((uint8_t)_cs.n8Secs10);
    LCDDigit6Set((uint8_t)_cs.n8Secs1);
}

int main(void) 
{
    CPUInit();
    I2CInit();
    
    LPDelayMS(50); // Low power delay, wait for buses to settle
    
    RTCInit();
    T1OscInit(); // Enable T1CKI, it's required for LCD, & is wired the RTC's CLKOUT running at 2048Hz
    LCDInit();
    LCDUpdate();
    T1Init(); // Setup timer 1 for 1 second interrupt: T1CKI should already have 2048Hz on it

    INTCONbits.PEIE=1;
    
    while (1)
    {
        LATEbits.LATE4=0; // GPIO for debug

        SLEEP();

        LATEbits.LATE4=1; // GPIO for debug

        TMR1-=TIMER1_DELAY;
        PIR1bits.TMR1IF=0;
       
//        --_cs.n8HalfSecond; // debug
        if (++_cs.n8HalfSecond>1)
        {
            _cs.n8HalfSecond=0;
            LCD_DATA_COL=1;
            if (++_cs.n8Secs1>9)
            {
                _cs.n8Secs1=0;
                if (++_cs.n8Secs10>5)
                {
                    _cs.n8Secs10=0;
                    if (++_cs.n8Mins1>9)
                    {
                        _cs.n8Mins1=0;
                        if (++_cs.n8Mins10>5)
                        {
                            _cs.n8Mins10=0;

                            if (++_cs.n8Hours1>9)
                            {
                                _cs.n8Hours1=0;
                                if (++_cs.n8Hours10>9)
                                {
                                    _cs.n8Hours10=0;
                                }
                                LCDDigit1Set((uint8_t)_cs.n8Hours10);
                            }
                            if (_cs.n8Hours1>3 && _cs.n8Hours10>=2)
                            {
                                _cs.n8Hours1=0;
                                _cs.n8Hours10=0;
                                LCDDigit1Set((uint8_t)_cs.n8Hours10);
                            }
                            LCDDigit2Set((uint8_t)_cs.n8Hours1);
                        }
                        LCDDigit3Set((uint8_t)_cs.n8Mins10);
                    }
                    LCDDigit4Set((uint8_t)_cs.n8Mins1);
                }
                LCDDigit5Set((uint8_t)_cs.n8Secs10);
            }
            LCDDigit6Set((uint8_t)_cs.n8Secs1);
        }
        else
        {
            LCD_DATA_COL=0;
        }
    }

    return 0;
}
